
/*
 * engine-3d v0.13.0
 * (c) 2018 @cocos
 * Released under the MIT License.
 */

'use strict';

// reference: https://github.com/mziccard/node-timsort

/**
 * Default minimum size of a run.
 */
var DEFAULT_MIN_MERGE = 32;

/**
 * Minimum ordered subsequece required to do galloping.
 */
var DEFAULT_MIN_GALLOPING = 7;

/**
 * Default tmp storage length. Can increase depending on the size of the
 * smallest run to merge.
 */
var DEFAULT_TMP_STORAGE_LENGTH = 256;

/**
 * Pre-computed powers of 10 for efficient lexicographic comparison of
 * small integers.
 */
var POWERS_OF_TEN = [1e0, 1e1, 1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8, 1e9];

/**
 * Estimate the logarithm base 10 of a small integer.
 *
 * @param {number} x - The integer to estimate the logarithm of.
 * @return {number} - The estimated logarithm of the integer.
 */
function log10(x) {
  if (x < 1e5) {
    if (x < 1e2) {
      return x < 1e1 ? 0 : 1;
    }

    if (x < 1e4) {
      return x < 1e3 ? 2 : 3;
    }

    return 4;
  }

  if (x < 1e7) {
    return x < 1e6 ? 5 : 6;
  }

  if (x < 1e9) {
    return x < 1e8 ? 7 : 8;
  }

  return 9;
}

/**
 * Default alphabetical comparison of items.
 *
 * @param {string|object|number} a - First element to compare.
 * @param {string|object|number} b - Second element to compare.
 * @return {number} - A positive number if a.toString() > b.toString(), a
 * negative number if .toString() < b.toString(), 0 otherwise.
 */
function alphabeticalCompare(a, b) {
  if (a === b) {
    return 0;
  }

  if (~~a === a && ~~b === b) {
    if (a === 0 || b === 0) {
      return a < b ? -1 : 1;
    }

    if (a < 0 || b < 0) {
      if (b >= 0) {
        return -1;
      }

      if (a >= 0) {
        return 1;
      }

      a = -a;
      b = -b;
    }

    var al = log10(a);
    var bl = log10(b);

    var t = 0;

    if (al < bl) {
      a *= POWERS_OF_TEN[bl - al - 1];
      b /= 10;
      t = -1;
    } else if (al > bl) {
      b *= POWERS_OF_TEN[al - bl - 1];
      a /= 10;
      t = 1;
    }

    if (a === b) {
      return t;
    }

    return a < b ? -1 : 1;
  }

  var aStr = String(a);
  var bStr = String(b);

  if (aStr === bStr) {
    return 0;
  }

  return aStr < bStr ? -1 : 1;
}

/**
 * Compute minimum run length for TimSort
 *
 * @param {number} n - The size of the array to sort.
 */
function minRunLength(n) {
  var r = 0;

  while (n >= DEFAULT_MIN_MERGE) {
    r |= (n & 1);
    n >>= 1;
  }

  return n + r;
}

/**
 * Counts the length of a monotonically ascending or strictly monotonically
 * descending sequence (run) starting at array[lo] in the range [lo, hi). If
 * the run is descending it is made ascending.
 *
 * @param {array} array - The array to reverse.
 * @param {number} lo - First element in the range (inclusive).
 * @param {number} hi - Last element in the range.
 * @param {function} compare - Item comparison function.
 * @return {number} - The length of the run.
 */
function makeAscendingRun(array, lo, hi, compare) {
  var runHi = lo + 1;

  if (runHi === hi) {
    return 1;
  }

  // Descending
  if (compare(array[runHi++], array[lo]) < 0) {
    while (runHi < hi && compare(array[runHi], array[runHi - 1]) < 0) {
      runHi++;
    }

    reverseRun(array, lo, runHi);
    // Ascending
  } else {
    while (runHi < hi && compare(array[runHi], array[runHi - 1]) >= 0) {
      runHi++;
    }
  }

  return runHi - lo;
}

/**
 * Reverse an array in the range [lo, hi).
 *
 * @param {array} array - The array to reverse.
 * @param {number} lo - First element in the range (inclusive).
 * @param {number} hi - Last element in the range.
 */
function reverseRun(array, lo, hi) {
  hi--;

  while (lo < hi) {
    var t = array[lo];
    array[lo++] = array[hi];
    array[hi--] = t;
  }
}

/**
 * Perform the binary sort of the array in the range [lo, hi) where start is
 * the first element possibly out of order.
 *
 * @param {array} array - The array to sort.
 * @param {number} lo - First element in the range (inclusive).
 * @param {number} hi - Last element in the range.
 * @param {number} start - First element possibly out of order.
 * @param {function} compare - Item comparison function.
 */
function binaryInsertionSort(array, lo, hi, start, compare) {
  if (start === lo) {
    start++;
  }

  for (; start < hi; start++) {
    var pivot = array[start];

    // Ranges of the array where pivot belongs
    var left = lo;
    var right = start;

    /*
     *   pivot >= array[i] for i in [lo, left)
     *   pivot <  array[i] for i in  in [right, start)
     */
    while (left < right) {
      var mid = (left + right) >>> 1;

      if (compare(pivot, array[mid]) < 0) {
        right = mid;
      } else {
        left = mid + 1;
      }
    }

    /*
     * Move elements right to make room for the pivot. If there are elements
     * equal to pivot, left points to the first slot after them: this is also
     * a reason for which TimSort is stable
     */
    var n = start - left;
    // Switch is just an optimization for small arrays
    switch (n) {
      case 3:
        array[left + 3] = array[left + 2];
      /* falls through */
      case 2:
        array[left + 2] = array[left + 1];
      /* falls through */
      case 1:
        array[left + 1] = array[left];
        break;
      default:
        while (n > 0) {
          array[left + n] = array[left + n - 1];
          n--;
        }
    }

    array[left] = pivot;
  }
}

/**
 * Find the position at which to insert a value in a sorted range. If the range
 * contains elements equal to the value the leftmost element index is returned
 * (for stability).
 *
 * @param {number} value - Value to insert.
 * @param {array} array - The array in which to insert value.
 * @param {number} start - First element in the range.
 * @param {number} length - Length of the range.
 * @param {number} hint - The index at which to begin the search.
 * @param {function} compare - Item comparison function.
 * @return {number} - The index where to insert value.
 */
function gallopLeft(value, array, start, length, hint, compare) {
  var lastOffset = 0;
  var maxOffset = 0;
  var offset = 1;

  if (compare(value, array[start + hint]) > 0) {
    maxOffset = length - hint;

    while (offset < maxOffset && compare(value, array[start + hint + offset]) > 0) {
      lastOffset = offset;
      offset = (offset << 1) + 1;

      if (offset <= 0) {
        offset = maxOffset;
      }
    }

    if (offset > maxOffset) {
      offset = maxOffset;
    }

    // Make offsets relative to start
    lastOffset += hint;
    offset += hint;

    // value <= array[start + hint]
  } else {
    maxOffset = hint + 1;
    while (offset < maxOffset && compare(value, array[start + hint - offset]) <= 0) {
      lastOffset = offset;
      offset = (offset << 1) + 1;

      if (offset <= 0) {
        offset = maxOffset;
      }
    }
    if (offset > maxOffset) {
      offset = maxOffset;
    }

    // Make offsets relative to start
    var tmp = lastOffset;
    lastOffset = hint - offset;
    offset = hint - tmp;
  }

  /*
   * Now array[start+lastOffset] < value <= array[start+offset], so value
   * belongs somewhere in the range (start + lastOffset, start + offset]. Do a
   * binary search, with invariant array[start + lastOffset - 1] < value <=
   * array[start + offset].
   */
  lastOffset++;
  while (lastOffset < offset) {
    var m = lastOffset + ((offset - lastOffset) >>> 1);

    if (compare(value, array[start + m]) > 0) {
      lastOffset = m + 1;

    } else {
      offset = m;
    }
  }
  return offset;
}

/**
 * Find the position at which to insert a value in a sorted range. If the range
 * contains elements equal to the value the rightmost element index is returned
 * (for stability).
 *
 * @param {number} value - Value to insert.
 * @param {array} array - The array in which to insert value.
 * @param {number} start - First element in the range.
 * @param {number} length - Length of the range.
 * @param {number} hint - The index at which to begin the search.
 * @param {function} compare - Item comparison function.
 * @return {number} - The index where to insert value.
 */
function gallopRight(value, array, start, length, hint, compare) {
  var lastOffset = 0;
  var maxOffset = 0;
  var offset = 1;

  if (compare(value, array[start + hint]) < 0) {
    maxOffset = hint + 1;

    while (offset < maxOffset && compare(value, array[start + hint - offset]) < 0) {
      lastOffset = offset;
      offset = (offset << 1) + 1;

      if (offset <= 0) {
        offset = maxOffset;
      }
    }

    if (offset > maxOffset) {
      offset = maxOffset;
    }

    // Make offsets relative to start
    var tmp = lastOffset;
    lastOffset = hint - offset;
    offset = hint - tmp;

    // value >= array[start + hint]
  } else {
    maxOffset = length - hint;

    while (offset < maxOffset && compare(value, array[start + hint + offset]) >= 0) {
      lastOffset = offset;
      offset = (offset << 1) + 1;

      if (offset <= 0) {
        offset = maxOffset;
      }
    }

    if (offset > maxOffset) {
      offset = maxOffset;
    }

    // Make offsets relative to start
    lastOffset += hint;
    offset += hint;
  }

  /*
   * Now array[start+lastOffset] < value <= array[start+offset], so value
   * belongs somewhere in the range (start + lastOffset, start + offset]. Do a
   * binary search, with invariant array[start + lastOffset - 1] < value <=
   * array[start + offset].
   */
  lastOffset++;

  while (lastOffset < offset) {
    var m = lastOffset + ((offset - lastOffset) >>> 1);

    if (compare(value, array[start + m]) < 0) {
      offset = m;

    } else {
      lastOffset = m + 1;
    }
  }

  return offset;
}

var TimSort = function TimSort(array, compare) {
  this.array = array;
  this.compare = compare;
  this.minGallop = DEFAULT_MIN_GALLOPING;
  this.length = array.length;

  this.tmpStorageLength = DEFAULT_TMP_STORAGE_LENGTH;
  if (this.length < 2 * DEFAULT_TMP_STORAGE_LENGTH) {
    this.tmpStorageLength = this.length >>> 1;
  }

  this.tmp = new Array(this.tmpStorageLength);

  this.stackLength =
    (this.length < 120 ? 5 :
      this.length < 1542 ? 10 :
        this.length < 119151 ? 19 : 40);

  this.runStart = new Array(this.stackLength);
  this.runLength = new Array(this.stackLength);
  this.stackSize = 0;
};

/**
 * Push a new run on TimSort's stack.
 *
 * @param {number} runStart - Start index of the run in the original array.
 * @param {number} runLength - Length of the run;
 */
TimSort.prototype.pushRun = function pushRun (runStart, runLength) {
  this.runStart[this.stackSize] = runStart;
  this.runLength[this.stackSize] = runLength;
  this.stackSize += 1;
};

/**
 * Merge runs on TimSort's stack so that the following holds for all i:
 * 1) runLength[i - 3] > runLength[i - 2] + runLength[i - 1]
 * 2) runLength[i - 2] > runLength[i - 1]
 */
TimSort.prototype.mergeRuns = function mergeRuns () {
    var this$1 = this;

  while (this.stackSize > 1) {
    var n = this$1.stackSize - 2;

    if ((n >= 1 &&
      this$1.runLength[n - 1] <= this$1.runLength[n] + this$1.runLength[n + 1]) ||
      (n >= 2 &&
      this$1.runLength[n - 2] <= this$1.runLength[n] + this$1.runLength[n - 1])) {

      if (this$1.runLength[n - 1] < this$1.runLength[n + 1]) {
        n--;
      }

    } else if (this$1.runLength[n] > this$1.runLength[n + 1]) {
      break;
    }
    this$1.mergeAt(n);
  }
};

/**
 * Merge all runs on TimSort's stack until only one remains.
 */
TimSort.prototype.forceMergeRuns = function forceMergeRuns () {
    var this$1 = this;

  while (this.stackSize > 1) {
    var n = this$1.stackSize - 2;

    if (n > 0 && this$1.runLength[n - 1] < this$1.runLength[n + 1]) {
      n--;
    }

    this$1.mergeAt(n);
  }
};

/**
 * Merge the runs on the stack at positions i and i+1. Must be always be called
 * with i=stackSize-2 or i=stackSize-3 (that is, we merge on top of the stack).
 *
 * @param {number} i - Index of the run to merge in TimSort's stack.
 */
TimSort.prototype.mergeAt = function mergeAt (i) {
  var compare = this.compare;
  var array = this.array;

  var start1 = this.runStart[i];
  var length1 = this.runLength[i];
  var start2 = this.runStart[i + 1];
  var length2 = this.runLength[i + 1];

  this.runLength[i] = length1 + length2;

  if (i === this.stackSize - 3) {
    this.runStart[i + 1] = this.runStart[i + 2];
    this.runLength[i + 1] = this.runLength[i + 2];
  }

  this.stackSize--;

  /*
   * Find where the first element in the second run goes in run1. Previous
   * elements in run1 are already in place
   */
  var k = gallopRight(array[start2], array, start1, length1, 0, compare);
  start1 += k;
  length1 -= k;

  if (length1 === 0) {
    return;
  }

  /*
   * Find where the last element in the first run goes in run2. Next elements
   * in run2 are already in place
   */
  length2 = gallopLeft(array[start1 + length1 - 1], array, start2, length2, length2 - 1, compare);

  if (length2 === 0) {
    return;
  }

  /*
   * Merge remaining runs. A tmp array with length = min(length1, length2) is
   * used
   */
  if (length1 <= length2) {
    this.mergeLow(start1, length1, start2, length2);

  } else {
    this.mergeHigh(start1, length1, start2, length2);
  }
};

/**
 * Merge two adjacent runs in a stable way. The runs must be such that the
 * first element of run1 is bigger than the first element in run2 and the
 * last element of run1 is greater than all the elements in run2.
 * The method should be called when run1.length <= run2.length as it uses
 * TimSort temporary array to store run1. Use mergeHigh if run1.length >
 * run2.length.
 *
 * @param {number} start1 - First element in run1.
 * @param {number} length1 - Length of run1.
 * @param {number} start2 - First element in run2.
 * @param {number} length2 - Length of run2.
 */
TimSort.prototype.mergeLow = function mergeLow (start1, length1, start2, length2) {

  var compare = this.compare;
  var array = this.array;
  var tmp = this.tmp;
  var i = 0;

  for (i = 0; i < length1; i++) {
    tmp[i] = array[start1 + i];
  }

  var cursor1 = 0;
  var cursor2 = start2;
  var dest = start1;

  array[dest++] = array[cursor2++];

  if (--length2 === 0) {
    for (i = 0; i < length1; i++) {
      array[dest + i] = tmp[cursor1 + i];
    }
    return;
  }

  if (length1 === 1) {
    for (i = 0; i < length2; i++) {
      array[dest + i] = array[cursor2 + i];
    }
    array[dest + length2] = tmp[cursor1];
    return;
  }

  var minGallop = this.minGallop;

  while (true) {
    var count1 = 0;
    var count2 = 0;
    var exit = false;

    do {
      if (compare(array[cursor2], tmp[cursor1]) < 0) {
        array[dest++] = array[cursor2++];
        count2++;
        count1 = 0;

        if (--length2 === 0) {
          exit = true;
          break;
        }

      } else {
        array[dest++] = tmp[cursor1++];
        count1++;
        count2 = 0;
        if (--length1 === 1) {
          exit = true;
          break;
        }
      }
    } while ((count1 | count2) < minGallop);

    if (exit) {
      break;
    }

    do {
      count1 = gallopRight(array[cursor2], tmp, cursor1, length1, 0, compare);

      if (count1 !== 0) {
        for (i = 0; i < count1; i++) {
          array[dest + i] = tmp[cursor1 + i];
        }

        dest += count1;
        cursor1 += count1;
        length1 -= count1;
        if (length1 <= 1) {
          exit = true;
          break;
        }
      }

      array[dest++] = array[cursor2++];

      if (--length2 === 0) {
        exit = true;
        break;
      }

      count2 = gallopLeft(tmp[cursor1], array, cursor2, length2, 0, compare);

      if (count2 !== 0) {
        for (i = 0; i < count2; i++) {
          array[dest + i] = array[cursor2 + i];
        }

        dest += count2;
        cursor2 += count2;
        length2 -= count2;

        if (length2 === 0) {
          exit = true;
          break;
        }
      }
      array[dest++] = tmp[cursor1++];

      if (--length1 === 1) {
        exit = true;
        break;
      }

      minGallop--;

    } while (count1 >= DEFAULT_MIN_GALLOPING || count2 >= DEFAULT_MIN_GALLOPING);

    if (exit) {
      break;
    }

    if (minGallop < 0) {
      minGallop = 0;
    }

    minGallop += 2;
  }

  this.minGallop = minGallop;

  if (minGallop < 1) {
    this.minGallop = 1;
  }

  if (length1 === 1) {
    for (i = 0; i < length2; i++) {
      array[dest + i] = array[cursor2 + i];
    }
    array[dest + length2] = tmp[cursor1];

  } else if (length1 === 0) {
    throw new Error('mergeLow preconditions were not respected');

  } else {
    for (i = 0; i < length1; i++) {
      array[dest + i] = tmp[cursor1 + i];
    }
  }
};

/**
 * Merge two adjacent runs in a stable way. The runs must be such that the
 * first element of run1 is bigger than the first element in run2 and the
 * last element of run1 is greater than all the elements in run2.
 * The method should be called when run1.length > run2.length as it uses
 * TimSort temporary array to store run2. Use mergeLow if run1.length <=
 * run2.length.
 *
 * @param {number} start1 - First element in run1.
 * @param {number} length1 - Length of run1.
 * @param {number} start2 - First element in run2.
 * @param {number} length2 - Length of run2.
 */
TimSort.prototype.mergeHigh = function mergeHigh (start1, length1, start2, length2) {
  var compare = this.compare;
  var array = this.array;
  var tmp = this.tmp;
  var i = 0;

  for (i = 0; i < length2; i++) {
    tmp[i] = array[start2 + i];
  }

  var cursor1 = start1 + length1 - 1;
  var cursor2 = length2 - 1;
  var dest = start2 + length2 - 1;
  var customCursor = 0;
  var customDest = 0;

  array[dest--] = array[cursor1--];

  if (--length1 === 0) {
    customCursor = dest - (length2 - 1);

    for (i = 0; i < length2; i++) {
      array[customCursor + i] = tmp[i];
    }

    return;
  }

  if (length2 === 1) {
    dest -= length1;
    cursor1 -= length1;
    customDest = dest + 1;
    customCursor = cursor1 + 1;

    for (i = length1 - 1; i >= 0; i--) {
      array[customDest + i] = array[customCursor + i];
    }

    array[dest] = tmp[cursor2];
    return;
  }

  var minGallop = this.minGallop;

  while (true) {
    var count1 = 0;
    var count2 = 0;
    var exit = false;

    do {
      if (compare(tmp[cursor2], array[cursor1]) < 0) {
        array[dest--] = array[cursor1--];
        count1++;
        count2 = 0;
        if (--length1 === 0) {
          exit = true;
          break;
        }

      } else {
        array[dest--] = tmp[cursor2--];
        count2++;
        count1 = 0;
        if (--length2 === 1) {
          exit = true;
          break;
        }
      }

    } while ((count1 | count2) < minGallop);

    if (exit) {
      break;
    }

    do {
      count1 = length1 - gallopRight(tmp[cursor2], array, start1, length1, length1 - 1, compare);

      if (count1 !== 0) {
        dest -= count1;
        cursor1 -= count1;
        length1 -= count1;
        customDest = dest + 1;
        customCursor = cursor1 + 1;

        for (i = count1 - 1; i >= 0; i--) {
          array[customDest + i] = array[customCursor + i];
        }

        if (length1 === 0) {
          exit = true;
          break;
        }
      }

      array[dest--] = tmp[cursor2--];

      if (--length2 === 1) {
        exit = true;
        break;
      }

      count2 = length2 - gallopLeft(array[cursor1], tmp, 0, length2, length2 - 1, compare);

      if (count2 !== 0) {
        dest -= count2;
        cursor2 -= count2;
        length2 -= count2;
        customDest = dest + 1;
        customCursor = cursor2 + 1;

        for (i = 0; i < count2; i++) {
          array[customDest + i] = tmp[customCursor + i];
        }

        if (length2 <= 1) {
          exit = true;
          break;
        }
      }

      array[dest--] = array[cursor1--];

      if (--length1 === 0) {
        exit = true;
        break;
      }

      minGallop--;

    } while (count1 >= DEFAULT_MIN_GALLOPING || count2 >= DEFAULT_MIN_GALLOPING);

    if (exit) {
      break;
    }

    if (minGallop < 0) {
      minGallop = 0;
    }

    minGallop += 2;
  }

  this.minGallop = minGallop;

  if (minGallop < 1) {
    this.minGallop = 1;
  }

  if (length2 === 1) {
    dest -= length1;
    cursor1 -= length1;
    customDest = dest + 1;
    customCursor = cursor1 + 1;

    for (i = length1 - 1; i >= 0; i--) {
      array[customDest + i] = array[customCursor + i];
    }

    array[dest] = tmp[cursor2];

  } else if (length2 === 0) {
    throw new Error('mergeHigh preconditions were not respected');

  } else {
    customCursor = dest - (length2 - 1);
    for (i = 0; i < length2; i++) {
      array[customCursor + i] = tmp[i];
    }
  }
};

/**
 * Sort an array in the range [lo, hi) using TimSort.
 *
 * @param {array} array - The array to sort.
 * @param {number} lo - First element in the range (inclusive).
 * @param {number} hi - Last element in the range.
 * @param {function=} compare - Item comparison function. Default is alphabetical.
 */
function sort (array, lo, hi, compare) {
  if (!Array.isArray(array)) {
    throw new TypeError('Can only sort arrays');
  }

  /*
   * Handle the case where a comparison function is not provided. We do
   * lexicographic sorting
   */

  if (lo === undefined) {
    lo = 0;
  }

  if (hi === undefined) {
    hi = array.length;
  }

  if (compare === undefined) {
    compare = alphabeticalCompare;
  }

  var remaining = hi - lo;

  // The array is already sorted
  if (remaining < 2) {
    return;
  }

  var runLength = 0;
  // On small arrays binary sort can be used directly
  if (remaining < DEFAULT_MIN_MERGE) {
    runLength = makeAscendingRun(array, lo, hi, compare);
    binaryInsertionSort(array, lo, hi, lo + runLength, compare);
    return;
  }

  var ts = new TimSort(array, compare);

  var minRun = minRunLength(remaining);

  do {
    runLength = makeAscendingRun(array, lo, hi, compare);
    if (runLength < minRun) {
      var force = remaining;
      if (force > minRun) {
        force = minRun;
      }

      binaryInsertionSort(array, lo, lo + force, lo + runLength, compare);
      runLength = force;
    }
    // Push new run and merge if necessary
    ts.pushRun(lo, runLength);
    ts.mergeRuns();

    // Go find next run
    remaining -= runLength;
    lo += runLength;

  } while (remaining !== 0);

  // Force merging of remaining runs
  ts.forceMergeRuns();
}

var FixedArray = function FixedArray(size) {
  this._count = 0;
  this._data = new Array(size);
};

var prototypeAccessors = { length: { configurable: true },data: { configurable: true } };

FixedArray.prototype._resize = function _resize (size) {
    var this$1 = this;

  if (size > this._data.length) {
    for (var i = this._data.length; i < size; ++i) {
      this$1._data[i] = undefined;
    }
  }
};

prototypeAccessors.length.get = function () {
  return this._count;
};

prototypeAccessors.data.get = function () {
  return this._data;
};

FixedArray.prototype.reset = function reset () {
    var this$1 = this;

  for (var i = 0; i < this._count; ++i) {
    this$1._data[i] = undefined;
  }

  this._count = 0;
};

FixedArray.prototype.push = function push (val) {
  if (this._count >= this._data.length) {
    this._resize(this._data.length * 2);
  }

  this._data[this._count] = val;
  ++this._count;
};

FixedArray.prototype.pop = function pop () {
  --this._count;

  if (this._count < 0) {
    this._count = 0;
  }

  var ret = this._data[this._count];
  this._data[this._count] = undefined;

  return ret;
};

FixedArray.prototype.fastRemove = function fastRemove (idx) {
  if (idx >= this._count) {
    return;
  }

  var last = this._count - 1;
  this._data[idx] = this._data[last];
  this._data[last] = undefined;
  this._count -= 1;
};

FixedArray.prototype.indexOf = function indexOf (val) {
  var idx = this._data.indexOf(val);
  if (idx >= this._count) {
    return -1;
  }

  return idx;
};

FixedArray.prototype.sort = function sort$1 (cmp) {
  return sort(this._data, 0, this._count, cmp);
};

Object.defineProperties( FixedArray.prototype, prototypeAccessors );

var Pool = function Pool(fn, size) {
  var this$1 = this;

  this._fn = fn;
  this._idx = size - 1;
  this._frees = new Array(size);

  for (var i = 0; i < size; ++i) {
    this$1._frees[i] = fn();
  }
};

Pool.prototype._expand = function _expand (size) {
    var this$1 = this;

  var old = this._frees;
  this._frees = new Array(size);

  var len = size - old.length;
  for (var i = 0; i < len; ++i) {
    this$1._frees[i] = this$1._fn();
  }

  for (var i$1 = len, j = 0; i$1 < size; ++i$1, ++j) {
    this$1._frees[i$1] = old[j];
  }

  this._idx += len;
};

Pool.prototype.alloc = function alloc () {
  // create some more space (expand by 20%, minimum 1)
  if (this._idx < 0) {
    this._expand(Math.round(this._frees.length * 1.2) + 1);
  }

  var ret = this._frees[this._idx];
  this._frees[this._idx] = null;
  --this._idx;

  return ret;
};

Pool.prototype.free = function free (obj) {
  ++this._idx;
  this._frees[this._idx] = obj;
};

// NOTE: you must have `_prev` and `_next` field in the object returns by `fn`

var LinkedArray = function LinkedArray(fn, size) {
  this._fn = fn;
  this._count = 0;
  this._head = null;
  this._tail = null;

  this._pool = new Pool(fn, size);
};

var prototypeAccessors$1 = { head: { configurable: true },tail: { configurable: true },length: { configurable: true } };

prototypeAccessors$1.head.get = function () {
  return this._head;
};

prototypeAccessors$1.tail.get = function () {
  return this._tail;
};

prototypeAccessors$1.length.get = function () {
  return this._count;
};

LinkedArray.prototype.add = function add () {
  var node = this._pool.alloc();

  if (!this._tail) {
    this._head = node;
  } else {
    this._tail._next = node;
    node._prev = this._tail;
  }
  this._tail = node;
  this._count += 1;

  return node;
};

LinkedArray.prototype.remove = function remove (node) {
  if (node._prev) {
    node._prev._next = node._next;
  } else {
    this._head = node._next;
  }

  if (node._next) {
    node._next._prev = node._prev;
  } else {
    this._tail = node._prev;
  }

  node._next = null;
  node._prev = null;
  this._pool.free(node);
  this._count -= 1;
};

LinkedArray.prototype.forEach = function forEach (fn, binder) {
    var this$1 = this;

  var cursor = this._head;
  if (!cursor) {
    return;
  }

  if (binder) {
    fn = fn.bind(binder);
  }

  var idx = 0;
  var next = cursor;

  while (cursor) {
    next = cursor._next;
    fn(cursor, idx, this$1);

    cursor = next;
    ++idx;
  }
};

Object.defineProperties( LinkedArray.prototype, prototypeAccessors$1 );

var RecyclePool = function RecyclePool(fn, size) {
  var this$1 = this;

  this._fn = fn;
  this._count = 0;
  this._data = new Array(size);

  for (var i = 0; i < size; ++i) {
    this$1._data[i] = fn();
  }
};

var prototypeAccessors$2 = { length: { configurable: true },data: { configurable: true } };

prototypeAccessors$2.length.get = function () {
  return this._count;
};

prototypeAccessors$2.data.get = function () {
  return this._data;
};

RecyclePool.prototype.reset = function reset () {
  this._count = 0;
};

RecyclePool.prototype.resize = function resize (size) {
    var this$1 = this;

  if (size > this._data.length) {
    for (var i = this._data.length; i < size; ++i) {
      this$1._data[i] = this$1._fn();
    }
  }
};

RecyclePool.prototype.add = function add () {
  if (this._count >= this._data.length) {
    this.resize(this._data.length * 2);
  }

  return this._data[this._count++];
};

RecyclePool.prototype.remove = function remove (idx) {
  if (idx >= this._count) {
    return;
  }

  var last = this._count - 1;
  var tmp = this._data[idx];
  this._data[idx] = this._data[last];
  this._data[last] = tmp;
  this._count -= 1;
};

RecyclePool.prototype.sort = function sort$1 (cmp) {
  return sort(this._data, 0, this._count, cmp);
};

Object.defineProperties( RecyclePool.prototype, prototypeAccessors$2 );

var _bufferPools = Array(8);
for (var i = 0; i < 8; ++i) {
  _bufferPools[i] = [];
}

var KEY_NONE = 0;
var KEY_DOWN = 1;
var KEY_PRESSING = 2;
var KEY_UP = 3;

var TOUCH_START = 0;
var TOUCH_PRESSING = 1;
var TOUCH_END = 2;
var TOUCH_CANCEL = 3;

var _dragMask = null;
var _phases = [
  'start',
  'pressing',
  'end',
  'cancel'
];

var _states = [
  'none',
  'down',
  'pressing',
  'up' ];

var Input = function Input(element, opts) {
  var this$1 = this;

  opts = opts || {};

  if (!_dragMask && opts.useMask) {
    _dragMask = document.createElement('div');
    _dragMask.classList.add('drag-mask');
    _dragMask.style.position = 'fixed';
    _dragMask.style.zIndex = '9999';
    _dragMask.style.top = '0';
    _dragMask.style.right = '0';
    _dragMask.style.bottom = '0';
    _dragMask.style.left = '0';
    _dragMask.oncontextmenu = function () { return false; };
  }

  this._element = element || document.body;

  // setup options

  this._enabled = true;
  if (opts.enabled !== undefined) {
    this._enabled = opts.enabled;
  }

  this._lock = false;
  if (opts.lock !== undefined) {
    this._lock = opts.lock;
  }

  this._useMask = false;
  if (opts.useMask !== undefined) {
    this._useMask = opts.useMask;
  }

  this._maskCursor = 'default';
  if (opts.maskCursor !== undefined) {
    this._maskCursor = opts.maskCursor;
  }

  this._invertY = false;
  if (opts.invertY !== undefined) {
    this._invertY = opts.invertY;
  }

  //
  var ua = navigator.userAgent.toLowerCase();
  if (/mobile|android|iphone|ipad|phone/i.test(ua)) {
    this._hasTouch = true;
  }

  // mouse internal states
  this._globalEventInstalled = false;
  this._pointerLocked = false;
  this._mouseGrabbed = false;

  this._bcr = element.getBoundingClientRect();

  // the mouse state
  this._mouse = {
    x: 0,
    y: 0,
    dx: 0,
    dy: 0,
    prevX: 0,
    prevY: 0,

    // mouse wheel (delta)
    scrollX: 0,
    scrollY: 0,

    // buttons
    left: KEY_NONE,
    right: KEY_NONE,
    middle: KEY_NONE,
  };

  // the keyboard state
  this._keys = new LinkedArray(function () {
    return {
      _next: null,
      _prev: null,
      _state: 0,
      key: '',
      get state () {
        return _states[this._state];
      }
    };
  }, 100);

  //the touch state
  this._touches = new RecyclePool(function () {
    return {
      id: -1, // touch.identifier
      x: 0,
      y: 0,
      dx: 0,
      dy: 0,
      prevX: 0,
      prevY: 0,
      get phase() {
        return _phases[this._phase];
      },
      _phase: -1, // 0: START, 1: PRESSING, 2: END
    };
  }, 16);

  // mousemove
  this._mousemoveHandle = function (event) {
    event.preventDefault();
    event.stopPropagation();

    this$1._mouse.dx = event.movementX;
    this$1._mouse.dy = event.movementY;

    if (this$1._pointerLocked) {
      this$1._mouse.x += event.movementX;
      if (this$1._invertY) {
        this$1._mouse.y -= event.movementY;
      } else {
        this$1._mouse.y += event.movementY;
      }
    } else {
      this$1._mouse.x = this$1._calcOffsetX(event.clientX);
      this$1._mouse.y = this$1._calcOffsetY(event.clientY);
    }
  };

  // mousewheel
  this._mousewheelHandle = function (event) {
    event.preventDefault();
    event.stopPropagation();

    this$1._mouse.scrollX = event.deltaX;
    this$1._mouse.scrollY = event.deltaY;
  };

  this._domMouseWheelHandle = function (event) {
    event.preventDefault();
    event.stopPropagation();

    // No horizon scrolling current.
    this$1._mouse.scrollX = 0;
    // When scroll forward, it's multiply of -3,
    // otherwise when scroll backward, it's multiply of 3.
    this$1._mouse.scrollY = event.detail / 3;
  };

  // mousedown
  this._mousedownHandle = function (event) {
    // NOTE: this will prevent mouse enter the text selection state.
    event.preventDefault();
    event.stopPropagation();

    if (this$1._lock) {
      this$1._lockPointer(true);
    }

    this$1._installGlobalEvents();
    this$1._element.focus();

    // handle mouse button
    switch (event.button) {
      // left mouse down
      case 0:
        // NOTE: do not reset KEY_DOWN when it already pressed
        if (this$1._mouse.left !== KEY_PRESSING) {
          this$1._mouse.left = KEY_DOWN;
        }
        break;

      // middle mouse down
      case 1:
        // NOTE: do not reset KEY_DOWN when it already pressed
        if (this$1._mouse.middle !== KEY_PRESSING) {
          this$1._mouse.middle = KEY_DOWN;
        }
        break;

      // right mouse down
      case 2:
        // NOTE: do not reset KEY_DOWN when it already pressed
        if (this$1._mouse.right !== KEY_PRESSING) {
          this$1._mouse.right = KEY_DOWN;
        }
        break;
    }
  };

  // mouseup
  this._mouseupHandle = function (event) {
    event.preventDefault();
    event.stopPropagation();

    // reset mouse position
    this$1._mouse.dx = event.movementX;
    this$1._mouse.dy = event.movementY;
    this$1._mouse.prevX = this$1._mouse.x = this$1._calcOffsetX(event.clientX);
    this$1._mouse.prevY = this$1._mouse.y = this$1._calcOffsetY(event.clientY);

    // handle mouse button
    switch (event.button) {
      // left mouse up
      case 0:
        this$1._mouse.left = KEY_UP;
        break;

      // middle mouse up
      case 1:
        this$1._mouse.middle = KEY_UP;
        break;

      // right mouse up
      case 2:
        this$1._mouse.right = KEY_UP;
        break;
    }
  };

  // mouseenter
  this._mouseenterHandle = function (event) {
    event.preventDefault();
    event.stopPropagation();

    this$1._mouse.dx = 0.0;
    this$1._mouse.dy = 0.0;
    this$1._mouse.prevX = this$1._mouse.x = this$1._calcOffsetX(event.clientX);
    this$1._mouse.prevY = this$1._mouse.y = this$1._calcOffsetY(event.clientY);
  };

  // mouseleave
  this._mouseleaveHandle = function (event) {
    event.preventDefault();
    event.stopPropagation();

    if (this$1._mouseGrabbed) {
      return;
    }

    this$1._uninstallGlobalEvents();

    this$1._mouse.dx = event.movementX;
    this$1._mouse.dy = event.movementY;
    this$1._mouse.prevX = this$1._mouse.x = this$1._calcOffsetX(event.clientX);
    this$1._mouse.prevY = this$1._mouse.y = this$1._calcOffsetY(event.clientY);
  };

  // keydown
  this._keydownHandle = function (event) {
    event.stopPropagation();

    var iter = this$1._keys.head;
    while (iter) {
      if (iter.key === event.key) {
        break;
      }
      iter = iter._next;
    }

    // NOTE: do not reset KEY_DOWN when it already pressed
    if (iter && iter._state === KEY_PRESSING) {
      return;
    }

    if (!iter) {
      iter = this$1._keys.add();
    }
    iter.key = event.key;
    iter._state = KEY_DOWN;
  };

  // keyup
  this._keyupHandle = function (event) {
    event.stopPropagation();

    var iter = this$1._keys.head;
    while (iter) {
      if (iter.key === event.key) {
        break;
      }
      iter = iter._next;
    }

    if (iter) {
      this$1._keys.remove(iter);
    }

    iter = this$1._keys.add();
    iter.key = event.key;
    iter._state = KEY_UP;
  };

  // touchstart
  this._touchstartHandle = function (event) {
    event.preventDefault();

    for (var i = 0; i < event.changedTouches.length; i++) {
      var changedTouch = event.changedTouches[i];
      var touch = this$1._touches.add();

      touch.id = changedTouch.identifier;
      touch._phase = TOUCH_START;
      touch.x = this$1._calcOffsetX(changedTouch.clientX);
      touch.y = this$1._calcOffsetY(changedTouch.clientY);
      touch.dx = 0;
      touch.dy = 0;
      touch.prevX = 0;
      touch.prevY = 0;
    }
  };

  // touchmove
  this._touchmoveHandle = function (event) {
    event.preventDefault();

    for (var i = 0; i < this$1._touches.length; i++) {
      for (var j = 0; j < event.changedTouches.length; j++) {
        var touch = this$1._touches.data[i];
        var changedTouch = event.changedTouches[j];

        if (touch.id === changedTouch.identifier) {
          touch._phase = TOUCH_PRESSING;
          touch.x = this$1._calcOffsetX(changedTouch.clientX);
          touch.y = this$1._calcOffsetY(changedTouch.clientY);
          touch.dx = touch.x - touch.prevX;
          touch.dy = touch.y - touch.prevY;
        }
      }
    }
  };

  // touchend
  this._touchendHandle = function (event) {
    event.preventDefault();

    for (var i = 0; i < this$1._touches.length; i++) {
      for (var j = 0; j < event.changedTouches.length; j++) {
        var touch = this$1._touches.data[i];
        var changedTouch = event.changedTouches[j];

        if (touch.id === changedTouch.identifier) {
          touch._phase = TOUCH_END;
          touch.prevX = touch.x = this$1._calcOffsetX(changedTouch.clientX);
          touch.prevY = touch.y = this$1._calcOffsetY(changedTouch.clientY);
          touch.dx = 0;
          touch.dy = 0;
        }
      }
    }
  };

  // touchcancel
  this._touchcancelHandle = function (event) {
    event.preventDefault();

    for (var i = 0; i < this$1._touches.length; i++) {
      for (var j = 0; j < event.changedTouches.length; j++) {
        var touch = this$1._touches.data[i];
        var changedTouch = event.changedTouches[j];

        if (touch.id === changedTouch.identifier) {
          touch._phase = TOUCH_CANCEL;
          touch.prevX = touch.x = this$1._calcOffsetX(changedTouch.clientX);
          touch.prevY = touch.y = this$1._calcOffsetY(changedTouch.clientY);
          touch.dx = 0;
          touch.dy = 0;
        }
      }
    }
  };

  // contextmenu
  this._contextmenuHandle = function (event) {
    event.preventDefault();
    event.stopPropagation();
  };

  if (this._enabled) {
    this._registerEvents();
  }
};

var prototypeAccessors$3 = { enabled: { configurable: true },hasTouch: { configurable: true },mouseX: { configurable: true },mouseY: { configurable: true },mouseDeltaX: { configurable: true },mouseDeltaY: { configurable: true },mousePrevX: { configurable: true },mousePrevY: { configurable: true },mouseScrollX: { configurable: true },mouseScrollY: { configurable: true },mouseButtons: { configurable: true },touchCount: { configurable: true },hasKeyDown: { configurable: true },hasKeyUp: { configurable: true },hasMouseDown: { configurable: true },hasMouseUp: { configurable: true } };

/**
 * Firefox use 'DOMMouseScroll' as their mouse wheel event name.
 * Other browsers(Safari, Chrome, Edge have been tested) use 'mousewheel' instead.
*/

Input.prototype._registerMousewheelEvent = function _registerMousewheelEvent () {
  this._element.addEventListener('mousewheel', this._mousewheelHandle, { passive: false });
  this._element.addEventListener('DOMMouseScroll', this._domMouseWheelHandle, { passive: false });
};

Input.prototype._unregisterMousewheelEvent = function _unregisterMousewheelEvent () {
  this._element.removeEventListener('mousewheel', this._mousewheelHandle, { passive: true });
  this._element.removeEventListener('DOMMouseScroll', this._domMouseWheelHandle, { passive: true });
};

Input.prototype.destroy = function destroy () {
  this._element.removeEventListener('mousedown', this._mousedownHandle);
  this._element.removeEventListener('mouseenter', this._mouseenterHandle);
  this._element.removeEventListener('mouseleave', this._mouseleaveHandle);
  this._element.removeEventListener('mousemove', this._mousemoveHandle);
  this._unregisterMousewheelEvent();
  this._element.removeEventListener('keydown', this._keydownHandle);
  this._element.removeEventListener('keyup', this._keyupHandle);
  this._element.removeEventListener('touchstart', this._touchstartHandle);
  this._element.removeEventListener('touchend', this._touchendHandle);
  this._element.removeEventListener('touchcancel', this._touchcancelHandle);
  this._element.removeEventListener('touchmove', this._touchmoveHandle);

  this._element.removeEventListener('contextmenu', this._contextmenuHandle);

  this._uninstallGlobalEvents();
};

Input.prototype._registerEvents = function _registerEvents () {
  this._element.addEventListener('mousedown', this._mousedownHandle);
  this._element.addEventListener('mouseenter', this._mouseenterHandle);
  this._element.addEventListener('mouseleave', this._mouseleaveHandle);
  this._element.addEventListener('mousemove', this._mousemoveHandle);
  this._registerMousewheelEvent();
  this._element.addEventListener('keydown', this._keydownHandle);
  this._element.addEventListener('keyup', this._keyupHandle);
  this._element.addEventListener('touchstart', this._touchstartHandle, false);
  this._element.addEventListener('touchend', this._touchendHandle, false);
  this._element.addEventListener('touchcancel', this._touchcancelHandle, false);
  this._element.addEventListener('touchmove', this._touchmoveHandle, false);

  this._element.addEventListener('contextmenu', this._contextmenuHandle);
};

Input.prototype._installGlobalEvents = function _installGlobalEvents () {
  if (this._globalEventInstalled) {
    return;
  }

  document.addEventListener('mouseup', this._mouseupHandle);
  document.addEventListener('mousemove', this._mousemoveHandle);
  document.addEventListener('mousewheel', this._mousewheelHandle, { passive: true });

  if (this._useMask) {
    _dragMask.style.cursor = this._maskCursor || 'default';
    document.body.appendChild(_dragMask);
  }

  this._globalEventInstalled = true;
};

Input.prototype._uninstallGlobalEvents = function _uninstallGlobalEvents () {
  if (!this._globalEventInstalled) {
    return;
  }

  // if we have mouse key pressed, skip it
  if (
    (this._mouse.left !== KEY_NONE && this._mouse.left !== KEY_UP) ||
    (this._mouse.right !== KEY_NONE && this._mouse.right !== KEY_UP) ||
    (this._mouse.middle !== KEY_NONE && this._mouse.middle !== KEY_UP)
  ) {
    return;
  }

  // unlock mouse here
  this._lockPointer(false);

  // if we are grabbing mouse, skip it
  if (this._mouseGrabbed) {
    return;
  }

  document.removeEventListener('mouseup', this._mouseupHandle);
  document.removeEventListener('mousemove', this._mousemoveHandle);
  document.removeEventListener('mousewheel', this._mousewheelHandle, { passive: true });

  if (this._useMask) {
    _dragMask.remove();
  }

  this._globalEventInstalled = false;
};

// NOTE: in web-browser, requestPointerLock only works in mousedown event
Input.prototype._lockPointer = function _lockPointer (locked) {
  if (locked) {
    if (this._pointerLocked) {
      return;
    }

    if (this._element.requestPointerLock) {
      this._element.requestPointerLock();
      this._pointerLocked = true;
    }

    return;
  } else {
    if (!this._pointerLocked) {
      return;
    }

    if (document.exitPointerLock) {
      document.exitPointerLock();
      this._pointerLocked = false;
    }
  }
};

Input.prototype._calcOffsetX = function _calcOffsetX (clientX) {
  return clientX - this._bcr.left;
};

Input.prototype._calcOffsetY = function _calcOffsetY (clientY) {
  if (this._invertY) {
    return this._bcr.height - (clientY - this._bcr.top);
  }

  return clientY - this._bcr.top;
};

/**
 * @property {boolean} enabled
 */
prototypeAccessors$3.enabled.get = function () {
  return this._enabled;
};
prototypeAccessors$3.enabled.set = function (val) {
  if (this._enabled !== val) {
    this._enabled = val;

    if (this._enabled) {
      this._registerEvents();
      if (this._mouseGrabbed) {
        this._installGlobalEvents();
      }
    } else {
      this.destroy();
    }
  }
};

/**
 * @property {boolean} hasTouch
 */
prototypeAccessors$3.hasTouch.get = function () {
  return this._hasTouch;
};

/**
 * @property {number} mouseX
 */
prototypeAccessors$3.mouseX.get = function () {
  return this._mouse.x;
};

/**
 * @property {number} mouseY
 */
prototypeAccessors$3.mouseY.get = function () {
  return this._mouse.y;
};

/**
 * @property {number} mouseDeltaX
 */
prototypeAccessors$3.mouseDeltaX.get = function () {
  return this._mouse.dx;
};

/**
 * @property {number} mouseDeltaY
 */
prototypeAccessors$3.mouseDeltaY.get = function () {
  return this._mouse.dy;
};

/**
 * @property {number} mousePrevX
 */
prototypeAccessors$3.mousePrevX.get = function () {
  return this._mouse.prevX;
};

/**
 * @property {number} mousePrevY
 */
prototypeAccessors$3.mousePrevY.get = function () {
  return this._mouse.prevY;
};

/**
 * @property {number} mouseScrollX
 */
prototypeAccessors$3.mouseScrollX.get = function () {
  return this._mouse.scrollX;
};

/**
 * @property {number} mouseScrollY
 */
prototypeAccessors$3.mouseScrollY.get = function () {
  return this._mouse.scrollY;
};

/**
 * @property {number} mouseButtons - mouse buttons in pressing states
 */
prototypeAccessors$3.mouseButtons.get = function () {
  var buttons = 0;

  var btn = this._mouse.left;
  if (btn === KEY_DOWN || btn === KEY_PRESSING) {
    buttons |= 1;
  }

  btn = this._mouse.right;
  if (btn === KEY_DOWN || btn === KEY_PRESSING) {
    buttons |= 2;
  }

  btn = this._mouse.middle;
  if (btn === KEY_DOWN || btn === KEY_PRESSING) {
    buttons |= 4;
  }

  return buttons;
};

/**
 * @property {number} touchCount
 */
prototypeAccessors$3.touchCount.get = function () {
  return this._touches.length;
};

/**
 * @property {boolean} hasKeyDown
 */
prototypeAccessors$3.hasKeyDown.get = function () {
  var iter = this._keys.head;
  while (iter) {
    if (iter._state === KEY_DOWN) {
      return true;
    }
  }
  return false;
};

/**
 * @property {boolean} hasKeyUp
 */
prototypeAccessors$3.hasKeyUp.get = function () {
  var iter = this._keys.head;
  while (iter) {
    if (iter._state === KEY_UP) {
      return true;
    }
  }
  return false;
};

/**
 * @property {boolean} hasMouseDown
 */
prototypeAccessors$3.hasMouseDown.get = function () {
  if (
    this._mouse.left === KEY_DOWN ||
    this._mouse.middle === KEY_DOWN ||
    this._mouse.right === KEY_DOWN
  ) {
    return true;
  }

  return false;
};

/**
 * @property {boolean} hasMouseUp
 */
prototypeAccessors$3.hasMouseUp.get = function () {
  if (
    this._mouse.left === KEY_UP ||
    this._mouse.middle === KEY_UP ||
    this._mouse.right === KEY_UP
  ) {
    return true;
  }

  return false;
};

/**
 * @method getTouchInfo
 * @param {number} idx
 */
Input.prototype.getTouchInfo = function getTouchInfo (idx) {
  return this._touches.data[idx];
};

/**
 * @method reset
 *
 * Reset the input states.
 * NOTE: you should call this at the end of your frame.
 */
Input.prototype.reset = function reset () {
    var this$1 = this;

  if (this._enabled === false) {
    return;
  }

  // update mouse states
  this._mouse.prevX = this._mouse.x;
  this._mouse.prevY = this._mouse.y;

  this._mouse.dx = 0;
  this._mouse.dy = 0;

  this._mouse.scrollX = 0;
  this._mouse.scrollY = 0;

  if (this._mouse.left === KEY_DOWN) {
    this._mouse.left = KEY_PRESSING;
  } else if (this._mouse.left === KEY_UP) {
    this._mouse.left = KEY_NONE;
  }

  if (this._mouse.middle === KEY_DOWN) {
    this._mouse.middle = KEY_PRESSING;
  } else if (this._mouse.middle === KEY_UP) {
    this._mouse.middle = KEY_NONE;
  }

  if (this._mouse.right === KEY_DOWN) {
    this._mouse.right = KEY_PRESSING;
  } else if (this._mouse.right === KEY_UP) {
    this._mouse.right = KEY_NONE;
  }

  // update keyboard states
  var iter = this._keys.head;
  var next = iter;
  while (next) {
    iter = next;
    next = iter._next;

    if (iter._state === KEY_DOWN) {
      iter._state = KEY_PRESSING;
    } else if (iter._state === KEY_UP) {
      this$1._keys.remove(iter);
    }
  }

  // update touch states
  for (var i = 0; i < this._touches.length; i++) {
    this$1._touches.data[i].prevX = this$1._touches.data[i].x;
    this$1._touches.data[i].prevY = this$1._touches.data[i].y;
    this$1._touches.data[i].dx = 0;
    this$1._touches.data[i].dy = 0;
    if (this$1._touches.data[i]._phase === TOUCH_START) {
      this$1._touches.data[i]._phase = TOUCH_PRESSING;
    }
    if (this$1._touches.data[i]._phase === TOUCH_END || this$1._touches.data[i]._phase === TOUCH_CANCEL) {
      this$1._touches.remove(i);
    }
  }

  // check if uninstall global events
  this._uninstallGlobalEvents();
};

/**
 * @method resize
 *
 * Update cached bounding client size.
 */
Input.prototype.resize = function resize () {
  this._bcr = this._element.getBoundingClientRect();
};

/**
 * @method grabMouse
 * @param {boolean} grabbed
 *
 * Keep tracing mouse move event when mouse leave the target element.
 */
Input.prototype.grabMouse = function grabMouse (grabbed) {
  this._mouseGrabbed = grabbed;

  // NOTE: we can mark mouse grabbed, but don't register events for it.
  if (this._enabled === false) {
    return;
  }

  if (grabbed) {
    this._installGlobalEvents();
  } else {
    this._uninstallGlobalEvents();
  }
};

/**
 * @method mousedown
 * @param {string} name - 'left', 'right' or 'middle'
 */
Input.prototype.mousedown = function mousedown (name) {
  var btn = this._mouse[name];
  if (btn !== undefined) {
    return btn === KEY_DOWN;
  }

  return false;
};

/**
 * @method mousepress
 * @param {string} name - 'left', 'right' or 'middle'
 */
Input.prototype.mousepress = function mousepress (name) {
  var btn = this._mouse[name];
  if (btn !== undefined) {
    return btn === KEY_DOWN || btn === KEY_PRESSING;
  }

  return false;
};

/**
 * @method mouseup
 * @param {string} name - 'left', 'right' or 'middle'
 */
Input.prototype.mouseup = function mouseup (name) {
  var btn = this._mouse[name];
  if (btn !== undefined) {
    return btn === KEY_UP;
  }

  return false;
};

/**
 * @method keydown
 * @param {string} name
 */
Input.prototype.keydown = function keydown (name) {
  var iter = this._keys.head;
  while (iter) {
    if (iter.key === name && iter._state === KEY_DOWN) {
      return true;
    }
    iter = iter._next;
  }

  return false;
};

/**
 * @method keyup
 * @param {string} name
 */
Input.prototype.keyup = function keyup (name) {
  var iter = this._keys.head;
  while (iter) {
    if (iter.key === name && iter._state === KEY_UP) {
      return true;
    }
    iter = iter._next;
  }

  return false;
};

/**
 * @method keypress
 * @param {string} name
 */
Input.prototype.keypress = function keypress (name) {
  var iter = this._keys.head;
  while (iter) {
    if (iter.key === name &&
      (iter._state === KEY_DOWN || iter._state === KEY_PRESSING)
    ) {
      return true;
    }
    iter = iter._next;
  }

  return false;
};

Object.defineProperties( Input.prototype, prototypeAccessors$3 );

module.exports = Input;
//# sourceMappingURL=input.js.map
